#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
GeigerLog - A combination of data logger, data presenter, and data analyzer to
            handle Geiger counters as well as environmental sensors for
            Temperature, Pressure, Humidity, and else

Start as 'geigerlog -h' for help on available options and commands
Use document 'GeigerLog-Manual-v<version number>.pdf' for further details
"""


###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"


import sys                      # system functions

# Test for proper Python version
# Check must be done before importing from gsup_utils, as some imports fail on
# older version.
#   Python 3.5 reached the end of its life on September 13th, 2020.
#   Please upgrade your Python as Python 3.5 is no longer maintained.
#   pip 21.0 will drop support for Python 3.5 in January 2021.
# Python 3.5 does NOT support ordered dicts, resulting in major problems!
#
svi            = sys.version_info
currentVersion = "{}.{}.{}".format(svi[0], svi[1], svi[2])
msg            = ""
if svi < (3, 6):    # Python version >= v3.6 is required.
    msg += """
        This version of GeigerLog requires Python in version 3.6 or later!<br>
        Your Python version is: {}.<br>""".format(currentVersion)

    if svi < (3,) :
        msg += """
            Python 2 is deprecated since January 1, 2020. The preferred way is to
            upgrade to Python 3. You can download a new version of Python from:

            https://www.python.org/downloads/

            If you can't do that, you can download an old copy of
            GeigerLog which still runs on Python 2, from:

            https://sourceforge.net/projects/geigerlog/

            The last version of GeigerLog for Python 2 is GeigerLog 0.9.06.
            Future versions will all be for Python 3.
            """

        print(msg.replace("<br>", "\n"))
        print("\7")
        print("\a")
        sys.exit(1)

    elif svi < (3, 6):
        msg += """<br><br>
            You can download a new version of Python from:<br><br>

            <a href='https://www.python.org/downloads/'>https://www.python.org/downloads/</a>

            <br>Python version 3.9 is recommended.<br><br>
            Cannot continue, will exit.
            """

from   gsup_utils            import *

gglobs.python_failure = msg # the Python version failure message


def main():

    # set directories and file names
    gglobs.progName         = getProgName   ()
    gglobs.progPath         = getProgPath   ()
    gglobs.gresPath         = getGresPath   ()
    gglobs.dataPath         = getDataPath   ()
    gglobs.proglogPath      = getProglogPath()
    gglobs.stdlogPath       = getStdlogPath ()
    gglobs.configPath       = getConfigPath ()
    gglobs.fileDialogDir    = getDataPath   ()

# data directory: Make sure it exists; create it if needed
    # exit if it cannot be made or is not writable
    if os.access(gglobs.dataPath , os.F_OK): # test for file exists
        # dir exists, ok
        if not os.access(gglobs.dataPath , os.W_OK):
            # dir exists, but is not writable
            print("ERROR: main: Data directory '{}' exists, but is not writable".format(gglobs.dataDirectory))
            return 1
    else:
        # dir does not exist; make it
        try:
            os.mkdir(gglobs.dataPath )
        except Exception as e:
            # dir cannot be made
            print("ERROR: main: Could not make data directory '{}'".format(gglobs.dataDirectory))
            return 1

# gres directory: Make sure it exists and is readable
    if os.access(gglobs.gresPath , os.F_OK):
        # dir exists, ok
        if not os.access(gglobs.dataPath , os.R_OK):
            # dir exists, but is not readable
            print("ERROR: main: GeigerLog resource directory {} exists but is not readable".format(gglobs.gresPath))
            return 1
    else:
        print("ERROR: main: GeigerLog resource directory {} does not exist. You may have to reinstall GeigerLog".format(gglobs.gresPath))
        return 1

# parse command line options
    # sys.argv[0] is progname
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hdvwRVPs:", ["help", "debug", "verbose", "werbose", "Redirect", "Version", "Portlist", "style=", "hidpi="])
    except getopt.GetoptError as e :
        # print info like "option -a not recognized", then exit
        msg = "For Help use: './geigerlog -h'"
        exceptPrint(e, msg)
        return 1

    # processing the options
    for opt, optval in opts:
        if   opt in ("-h", "--help"):
            print (gglobs.helpOptions)
            return 0

        elif opt in ("-V", "--Version"):
            print("Version status:")
            for a in getVersionStatus(): print("   {:16s}: {}".format(a[0], a[1]))
            return 0

        elif opt in ("-P", "--Portlist"):   # which ports are available?
            lp = serial.tools.list_ports.comports(include_links=False)  # default; include_links=True also shows symlinks
            print("\nAvailable USB-to-Serial ports w/o symlinks:")
            for p in lp:     print("   ", p)
            if len(lp) == 0: print("   ", "None")
            return 0

        elif opt in ("-d", "--debug"):
            gglobs.debug    = True

        elif opt in ("-v", "--verbose"):
            gglobs.verbose  = True
            gglobs.debug    = True          # giving verbose implies debug

        elif opt in ("-w", "--werbose"):
            gglobs.werbose  = True
            gglobs.verbose  = True          # giving werbose implies verbose
            gglobs.debug    = True          # giving verbose implies debug

        elif opt in ("-R", "--Redirect"):
            gglobs.redirect = True


    # processing the args
    for arg in args:
        if arg == "showstyles":
            print("Styles found on system: ")
            for a in QStyleFactory.keys():  print("   " + a)
            return 0

        if arg == "devel":
            gglobs.devel    = True

        if arg == "devel1":
            gglobs.devel    = True
            gglobs.devel1   = True

        if arg == "devel2":
            gglobs.devel    = True
            gglobs.devel1   = True
            gglobs.devel2   = True

        if arg == "devel3":
            gglobs.devel    = True
            gglobs.devel1   = True
            gglobs.devel2   = True
            gglobs.devel3   = True

        if arg == "fullhist":   gglobs.fullhist         = True
        if arg == "keepFF":     gglobs.keepFF           = True

        if arg == "tput":       gglobs.tput             = True

        if arg == "testing":    gglobs.testing          = True
        if arg == "graphdemo":  gglobs.graphdemo        = True
        if arg == "stattest":   gglobs.stattest         = True
        if arg == "GStesting":  gglobs.GStesting        = True # required for GammaScout with simulation

        if arg == "test1":      gglobs.test1            = True
        if arg == "test2":      gglobs.test2            = True
        if arg == "test3":      gglobs.test3            = True
        if arg == "test4":      gglobs.test4            = True

        if arg == "gmc":        gglobs.GMCActivation    = True
        if arg == "audio":      gglobs.AudioActivation  = True
        if arg == "osterei":    gglobs.AudioOei         = True
        if arg == "radmon":     gglobs.RMActivation     = True
        if arg == "ambio":      gglobs.AmbioActivation  = True
        if arg == "scout":      gglobs.GSActivation     = True
        if arg == "i2c":        gglobs.I2CActivation    = True
        if arg == "labjack":    gglobs.LJActivation     = True
        if arg == "raspi":      gglobs.RaspiActivation  = True
        if arg == "simul":      gglobs.SimulActivation  = True
        if arg == "minimon":    gglobs.MiniMonActivation= True

# clear and initialize the program log file 'geigerlog.proglog'
    clearProgramLogFile()

# set terminal output with or without linebreaks (works in Linux only)
    executeTPUT(action = 'set')

# Signal handlers
    # Attention on Windows:
    # see: https://docs.microsoft.com/en-us/previous-versions/xdkz3x12(v=vs.140)
    # SIGINT :  CTRL+C signal, but: "SIGINT is not supported for any Win32 application."
    # SIGTSTP:  CTRL-Z signal, NOT supported on Windows!!!
    #
    #~print("signal.SIG_DFL: {:2d}  ".format(signal.SIG_DFL), signal.SIG_DFL) # 0
    #~print("signal.SIGINT:  {:2d}  ".format(signal.SIGINT),  signal.SIGINT ) # 2
    #~print("signal.SIGTERM: {:2d}  ".format(signal.SIGTERM), signal.SIGTERM) # 15
    #~print("signal.SIGTSTP: {:2d}  ".format(signal.SIGTSTP), signal.SIGTSTP) # 20

    #~print("SIGINT  handler before:    SIGINT : {:2d}, getsignal(SIGINT) : {}".format(signal.SIGINT,  signal.getsignal(signal.SIGINT )))
    #~print("SIGTSTP handler before:    SIGTSTP: {:2d}, getsignal(SIGTSTP): {}".format(signal.SIGTSTP, signal.getsignal(signal.SIGTSTP)))
    try:
        signal.signal(signal.SIGINT,  signal.SIG_DFL) # to allow shut down with ctrl-c when IN THE TERMINAL
        dprint("{:28s}: SIGINT : {:2d}, getsignal(SIGINT) : {}".format("SIGNAL SIGINT  set default", signal.SIGINT,  signal.getsignal(signal.SIGINT)))
    except Exception as e:
        exceptPrint(e, "SIGNAL SIGINT setting failed on platform: '{}'".format(platform.platform()))

    if "LINUX" in platform.platform().upper():
        try:
            signal.signal(signal.SIGTSTP, signal.SIG_IGN) # CTRL-Z signal will be ignored
            dprint("{:28s}: SIGTSTP: {:2d}, getsignal(SIGTSTP): {}".format("SIGNAL SIGTSTP set ignored", signal.SIGTSTP, signal.getsignal(signal.SIGTSTP)))
        except Exception as e:
            exceptPrint(e, "SIGNAL SIGTSTP setting failed on platform: '{}'".format(platform.platform()))
    #~print("SIGINT  handler after :    SIGINT:  {:2d}, getsignal(SIGINT) : {}".format(signal.SIGINT, signal.getsignal(signal.SIGINT)))
    #~print("SIGTSTP handler after :    SIGTSTP: {:2d}, getsignal(SIGTSTP): {}".format(signal.SIGTSTP, signal.getsignal(signal.SIGTSTP)))

    # test what the signals are connected to
    #~for i in range(0, 95): # fails with i=0 and i>64
        #~try:
            #~print(": i: {:3d}, signal.getsignal(i): {}".format(i, signal.getsignal(i)))
        #~except ValueError as e:
            #~print(": i: {:3d}, signal.getsignal(i): {}   ValueError".format(i, e))
        #~except Exception as e:
            #~print(": i: {:3d}, signal.getsignal(i): {}   Exception".format(i, e))

# Command line
    dprint("{:28s}: sys.argv: {}".format("Command line", sys.argv))
    dprint("{:28s}: options: {}, commands: {}".format("Options, Commands", opts, args))

# Platform, machine
    dprint("{:28s}: {}".format("Operating System", platform.platform()))
    dprint("{:28s}: {}, {}".format("Machine", platform.machine(), platform.architecture()[0]))

# Versions
    dprint("Version status: ")
    for a in getVersionStatus(): dprint("   {:25s}: {}".format(a[0], a[1]))

# Python Search Path
    dprint("Python Search Paths:")
    # sys.path.append('pfad/zu/irgend/was') # to extend the search path
    for a in sys.path:
        dprint("   {:25s}: {}".format("sys.path", a))

# Program Paths
    dprint("Program Paths:")
    dprint("   {:25s}: {}".format("progName",   gglobs.progName))
    dprint("   {:25s}: {}".format("progPath",   gglobs.progPath))
    dprint("   {:25s}: {}".format("gresPath",   gglobs.gresPath))
    dprint("   {:25s}: {}".format("dataPath",   gglobs.dataPath))
    dprint("   {:25s}: {}".format("proglogPath",gglobs.proglogPath))
    dprint("   {:25s}: {}".format("configPath", gglobs.configPath))

# Read the configuration file
    # - these settings will override defaults in gglobs.py
    # - command line options may override these settings
    readGeigerLogConfig()

###############################################################################
# Starting PyQt

# QT_STYLE_OVERRIDE
    # app = QApplication([]) -> results in message: (only PyQt5, not PyQt4!)
    #    QApplication: invalid style override passed, ignoring it.
    #            Available styles: Windows, Fusion
    # unless a '-style <valid style>' is in sys.argv.
    # app = QApplication([None, "-style", "Windows"])   # ok!
    # app = QApplication([None, "-style", "Fusion"])    # ok!
    # app = QApplication([None, "-style", ""])          # ok!
    #
    # This results from the setting of:  QT_STYLE_OVERRIDE="gtk" in the OS. (printenv | grep QT)
    # Can be removed on the shell:  export QT_STYLE_OVERRIDE="" -> then the message disappears
    # preferred: remove from within Python:
    os.environ["QT_STYLE_OVERRIDE"] = ""

# HiDPI screen
    # MUST be done before QApplication is started!
    # see my article: HOWTO - Using PyQt5 and matplotlib on HiDPI monitors (Python3).pdf
    #
    # QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    # !!!! im Raspi führte diese Anweisung zu einer Bildschirmgröße von 960x540,
    # !!!! obwohl Bildschirm auf 1920x1080 steht! -->unbrauchbar für Raspi
    #
    dprint("{:28s}".format("High DPI Scaling:"))
    if gglobs.hidpiActivation == "auto" or gglobs.hidpiActivation == "yes" :
        ehds = "EnableHighDpiScaling"
        uhdp = "UseHighDpiPixmaps"
        try:
            if hasattr(Qt, 'AA_EnableHighDpiScaling'):
                QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
                dprint("   {:25s}: {}".format(ehds, "Enabled"))
            else:
                edprint("   {:25s}: {}".format(ehds, "Attribute not available"))
        except Exception as e:
            edprint("   {:25s}: {} Exception: {}".format(ehds, "Failure!", e))

        try:
            if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
                QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
                dprint("   {:25s}: {}".format(uhdp, "Enabled"))
            else:
                edprint("   {:25s}: {}".format(uhdp, "Attribute not available"))
        except Exception as e:
            edprint("   {:25s}: {} Exception: {}".format(uhdp, "Failure!", e))
    else:
        dprint("   {:25s}: {}".format("High DPI Scaling", "Deselected"))

# Locale setting
    # if the system's local is used, then e.g. Germany's local is used
    # this makes the double validator use komma as period !!
    #
    # QLocale.OmitGroupSeparator        0x01
    # If this option is set, the number-to-string functions will not insert group
    # separators in their return values. The default is to insert group separators.
    # QLocale.RejectGroupSeparator  0x02
    # If this option is set, the string-to-number functions will fail if they
    # encounter group separators in their input. The default is to accept numbers
    # containing correctly placed group separators.
    mylocal = QLocale()
    mylocal.setDefault(QLocale(QLocale.English, QLocale.UnitedStates));
    mylocal.setNumberOptions(QLocale.OmitGroupSeparator | QLocale.RejectGroupSeparator);
    #print("QLocale.numberOptions (self): ", mylocal.numberOptions()) # is an object :-(


# QApplication
    # app = QApplication(sys.argv)   # conflicts in syntax with getopt (single vs double dash)
    gglobs.app = QApplication([])
    gglobs.app.setWindowIcon           (QIcon(QPixmap(os.path.join(gglobs.gresPath, 'icon_geigerlog.png')))) # sets the icon (overrun elsewhere)
    gglobs.app.setApplicationName      ("GeigerLog")               # sets the name 'GeigerLog' as the application name
    gglobs.app.setApplicationVersion   (gglobs.__version__)        # sets the GL version
    gglobs.app.setFont                 (QFont("Sans", 10))         # setting the application font size (default looks close to 11)

    #~# if the system's local is used, then e.g. Germany's local is used
    #~# this makes the double validator use komma as period !!
    #~mylocal = QLocale()
    #~mylocal.setDefault(QLocale(QLocale.English, QLocale.UnitedStates));
    #~#print("QLocale.numberOptions (self): ", mylocal.numberOptions()) # is an object :-(

    #~# QLocale.OmitGroupSeparator        0x01
    #~# If this option is set, the number-to-string functions will not insert group
    #~# separators in their return values. The default is to insert group separators.
    #~# QLocale.RejectGroupSeparator  0x02
    #~# If this option is set, the string-to-number functions will fail if they
    #~# encounter group separators in their input. The default is to accept numbers
    #~# containing correctly placed group separators.
    #~mylocal.setNumberOptions(QLocale.OmitGroupSeparator | QLocale.RejectGroupSeparator);

    # allows copy&Paste also on Win, but makes all Messageboxes as HTML coding
    # cumbersome with using space and tabs for formatting; do not use
    # gglobs.app.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")

    dprint("QCore Application:")
    dprint("   {:25s}: {}".format("applicationName()"     , QCoreApplication.applicationName()))
    dprint("   {:25s}: {}".format("applicationVersion()"  , QCoreApplication.applicationVersion()))
    dprint("   {:25s}: {}".format("applicationDirPath()"  , QCoreApplication.applicationDirPath()))
    dprint("   {:25s}: {}".format("applicationFilePath()" , QCoreApplication.applicationFilePath()))

# Library paths
    dprint("QCore Library paths:")
    #QCoreApplication.addLibraryPath(gglobs.progPath + "/custom_libs/") # will be listed only if folder exists
    #QCoreApplication.addLibraryPath("./");
    # needed? -> https://stackoverflow.com/questions/21268558/application-failed-to-start-because-it-could-not-find-or-load-the-qt-platform-pl
    libPaths    = QCoreApplication.libraryPaths()
    libPathStr  = "libraryPaths()"
    if len(libPaths) == 0: dprint("   {:25s}: {}".format(libPathStr, "No Library Paths"))
    else:
        for a in libPaths: dprint("   {:25s}: {}".format(libPathStr, a))

# QScreen
    # see: https://doc.qt.io/qt-5/qscreen.html#physicalDotsPerInch-prop
    # requires app.primaryScreen(), which is only available after app was started
    dprint("Connected Screen:")
    dprint("   {:25s}: {}".format("manufacturer",           QScreen.manufacturer(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("model       ",           QScreen.model(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("name        ",           QScreen.name(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("size        ",           QScreen.size(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("geometry    ",           QScreen.geometry(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("depth       ",           QScreen.depth(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("physicalSize",           QScreen.physicalSize(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("physicalDotsPerInch",    QScreen.physicalDotsPerInch(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("physicalDotsPerInchX",   QScreen.physicalDotsPerInchX(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("physicalDotsPerInchY",   QScreen.physicalDotsPerInchY(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("orientation ",           QScreen.orientation(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("devicePixelRatio",       QScreen.devicePixelRatio(gglobs.app.primaryScreen())))
    dprint("   {:25s}: {}".format("refreshRate ",           QScreen.refreshRate(gglobs.app.primaryScreen())))

# matplotlib HiDPI adjustment
    dprint("HiDPI MPL Scale:")
    if gglobs.hidpiScaleMPL == "auto":
        gglobs.hidpiScaleMPL = 100 / int(QScreen.devicePixelRatio(gglobs.app.primaryScreen()))
    dprint("   {:25s}: {}".format("HiDPI MPL Scale",      gglobs.hidpiScaleMPL))

# Style
    # latest news for PyQt5:
    # https://blog.qt.io/blog/2012/10/30/cleaning-up-styles-in-qt5-and-adding-fusion/
    # Summary: Default Qt5 has removed all styles except Windows (maybe) and added Fusion
    #
    # QStyleFactory::keys() returns a list of valid keys;
    # order of my preference; first is best: PyQt5: ['Fusion', 'Windows'] (PyQt5 has only 2)

    dprint("Styles:")
    dprint("   {:25s}: {}".format("Default",                gglobs.app.style().metaObject().className()))
    dprint("   {:25s}: {}".format("Styles found on system", QStyleFactory.keys()))
    for opt, optval in opts:
        if opt in ("-s", "--style"):
            available_styles = QStyleFactory.keys()
            if optval in available_styles:
                gglobs.windowStyle  = optval
                dprint("   {:25s}: '{}' - overrides default".format("Style per command line", gglobs.windowStyle))
            else:
                gglobs.windowStyle  = "auto"
                dprint("   {:25s}: '{}' - Style not available (check spelling, capitalization); ignoring".format("Style per command line", optval))

    if gglobs.windowStyle != "auto":    # existence of style had been verified
        result = gglobs.app.setStyle(gglobs.windowStyle)
        if result is None:  dprint("   Selected Style '{}' could not be applied".format(gglobs.windowStyle))
    dprint("   {:25s}: {}".format("Active Style is", gglobs.app.style().metaObject().className()))

# get the GUI
    from  ggeiger import ggeiger

# starting the GUI
    dprint(TGREEN + "Starting the GUI " + "-" * 110 + TDEFAULT)
    ex     = ggeiger()              # an instance of ggeiger; runs init and draws window
    status = gglobs.app.exec_()     # starts the QApplication([]); run the GUI until closure
    dprint(TGREEN + "Exited GUI with status of: {} ".format(status) + "-" * 97 + TDEFAULT)

# cleanup
    executeTPUT(action = 'clean')

    return status


if __name__ == '__main__':
    main()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
gdev_audio.py - GeigerLog commands to handle the audio clicks as CPM/CPS input
                using module sounddevice
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

"""
NOTES:
Raspi:      on Raspi sounddevice cannot be installed until libffi-dev is installed per 'sudo apt install libffi-dev'
Windows:    on Windows: make sure the soundsettings for input and output are set for 'standard' using Windows settings
Linux       with garbled sound and message in the terminal:
                "ALSA lib pcm.c:7963:(snd_pcm_recover) underrun occurred"
            execute 'pulseaudio -k' as regular user (not sudo!) this should solve it

            if not enough: look into '/etc/pulse/daemon.conf'. It may have lines:
                ; default-fragments = 4
                ; default-fragment-size-msec = 25
            uncommenting, and changing to:
                default-fragments = 5
                default-fragment-size-msec = 2
            might help
"""

# LATENCY:
#
# Suggested input/output latency in seconds.
#
# The special values 'low' and 'high' can be used to select the default
# low/high latency of the chosen device. 'high' is typically more robust
# (i.e. buffer under-/overflows are less likely), but the latency may be
# too large for interactive applications.
#
# NOTE: latencies are given in millisec!
# on my Desktop:
# 'default_high_input_latency':  0.034829931972789115   --> 35 ms
# 'default_low_input_latency'  : 0.008707482993197279   -->  9 ms
# 'default_high_output_latency': 0.034829931972789115   --> 35 ms
# 'default_low_output_latency':  0.008707482993197279   -->  9 ms
#
# on MSI (low power) computer: using Input: 'Mikrofon (2- High Definition Au'
# 'default_high_input_latency':  0.18                   --> 180 ms
# 'default_low_input_latency':   0.09                   -->  90 ms
# 'default_high_output_latency': 0.18                   --> 180 ms
# 'default_low_output_latency':  0.09                   -->  90 ms

# tested on MSI (low power) computer on Windows 10 :
#~sd.default.latency  = ('low', 'low')   --> results in poor, low counts
#~sd.default.latency  = ('high', 'high') --> results in higher counts, but still less than digital on GMC300E+
#~sd.default.latency  = (100, 100)       --> results in same counts as digital on GMC300E+
#~sd.default.latency  = 1000, or 10000 works, but no improvement
#~sd.default.latency  = (1, 1)           --> works just as well!!!
#~sd.default.latency  = (0.1, 0.1)       --> getting 'overflowed' !
#~sd.default.latency  = (0.3, 0.3)       --> getting some 'overflowed' !
#~sd.default.latency  = (0.5, 0.5)       --> works, no 'overflowed' observed over >30sec !

# tested on Raspi 4 with USB dongle "ID 1b3f:2008 Generalplus Technology Inc. "
# sd.default.latency  = (1.0, 1.0)       --> works, few 'overflowed' observed over >30sec !


from   gsup_utils       import *


def initAudioCounter():
    """Start the thread to monitor the audio pulses for clicks"""

    global AudioCounterThread

    fncname ="initAudioCounter: "

    dprint(fncname + "Initialzing AudioCounter")
    setDebugIndent(1)

    errmsg  = ""                            # what errors can be here? e.g no input device!
    gglobs.AudioDeviceName      = "AudioCounter"
    gglobs.AudioDeviceDetected  = gglobs.AudioDeviceName
    gglobs.Devices["Audio"][0]  = gglobs.AudioDeviceDetected

    if gglobs.AudioDevice       == "auto": gglobs.AudioDevice       = (None, None)      # the system's defaults
    if gglobs.AudioLatency      == "auto": gglobs.AudioLatency      = (1.0, 1.0)        # seems like a generic
    if gglobs.AudioPulseMax     == "auto": gglobs.AudioPulseMax     = 32768             # +/- 15 bit
    if gglobs.AudioPulseDir     == "auto": gglobs.AudioPulseDir     = False             # => negative
    if gglobs.AudioThreshold    == "auto": gglobs.AudioThreshold    = 60                # % of max
    if gglobs.AudioSensitivity  == "auto": gglobs.AudioSensitivity  = 154               # CPM/(µSv/h), = 0.065 µSv/h/CPM
    if gglobs.AudioVariables    == "auto": gglobs.AudioVariables    = "CPM3rd, CPS3rd"

    setCalibrations(gglobs.AudioVariables, gglobs.AudioSensitivity)

    setLoggableVariables("Audio", gglobs.AudioVariables)


#Sound Driver settings

    dprint(fncname + "Query Devices: \n",   sd.query_devices() )
    dprint(fncname + "Query Host Apis:")
    if gglobs.debug:
        for a in sd.query_hostapis(index=None): print(" ", a)

    # samplerate has only one single int value, but others
    # allow 2 values: 1st=input, 2nd=output
    sd.default.reset()                              # clear all settings to default
    sd.default.device       = gglobs.AudioDevice    # set the AudioDevice defaults
    sd.default.latency      = gglobs.AudioLatency   # set latency values in sec, like (1.0, 1.0)
    sd.default.dtype        = ('int16', 'int16')    # choose from: 'float32', 'int32', 'int16', 'int8', 'uint8'
    sd.default.channels     = (1, 1)                # options: 1, 2, more
    sd.default.samplerate   = 44100                 # options: 96000, 48000, 44100, 22050, 11025

    print()
    try:
        defaultDevice = sd.default.device
        dprint(fncname + "Default Device (Input, Output): {}".format(defaultDevice))
        if defaultDevice[1] != -1 :
            sdqd = sd.query_devices(device=None, kind='output')
            vprint(fncname  + "Output Device: ")
            for key in sdqd: vprint("   {:30s} : {}".format(key, sdqd[key]))
        else:
            edprint(fncname + "Output Device: ", "No Default Output Device!")

        if defaultDevice[0] != -1 :
            sdqd = sd.query_devices(device=None, kind='input')
            vprint(fncname  + "Input Device : ")
            for key in sdqd: vprint("   {:30s} : {}".format(key, sdqd[key]))
        else:
            edprint(fncname + "Input Device : ", "No Default Input Device!")

    except Exception as e:
        info = fncname + "Exception in querying default devices"
        exceptPrint(e, info)
    print()

    if defaultDevice[0] != -1:
        gglobs.AudioFormat      = sd.default.dtype
        gglobs.AudioChannels    = sd.default.channels
        gglobs.AudioRate        = sd.default.samplerate
        gglobs.AudioChunk       = 32

        dprint(fncname + "DEVICE:{}, CHANNELS:{}, FORMAT:{}, Latency:{}, Host API Index:{}, RATE:{}, CHUNK:{}"\
                        .format(
                                gglobs.AudioDevice,
                                gglobs.AudioChannels,
                                gglobs.AudioFormat,
                                gglobs.AudioLatency,
                                sd.default.hostapi,
                                gglobs.AudioRate,
                                gglobs.AudioChunk,
                               )
             )

        gglobs.AudioConnection = True

        AudioCounterThread = threading.Thread(target=AudioCounterThreadTarget, args=(None,))
        gglobs.AudioThreadStop = False
        AudioCounterThread.start()
        dprint(fncname + "Thread-status: is alive: ", AudioCounterThread.is_alive())

    else:
        errmsg  = "No Sound Input detected! Is there a proper connection?"

    setDebugIndent(0)

    return errmsg


def terminateAudioCounter():
    """Stop the thread to monitor the sounddev pulses for clicks"""

    global AudioCounterThread

    fncname ="terminateAudioCounter: "
    dprint(fncname)
    if not gglobs.AudioConnection: return fncname + "No AudioCounter connection"

    setDebugIndent(1)
    dprint(fncname + "stopping thread")
    gglobs.AudioThreadStop = True
    AudioCounterThread.join() # "This blocks the calling thread until the thread
                              #  whose join() method is called is terminated."

    # verify that thread has ended, but wait not longer than 5 sec (takes 0.006...0.016 ms)
    start = time.time()
    while AudioCounterThread.is_alive() and (time.time() - start) < 5:
        pass

    dprint(fncname + "thread-status: is alive: {}, waiting took:{:0.3f}ms".format(AudioCounterThread.is_alive(), 1000 * (time.time() - start)))
    gglobs.AudioConnection = False

    setDebugIndent(0)


def AudioCounterThreadTarget(Dummy):
    """The thread to read the sounddev input"""

    fncname = "AudioCounterThreadTarget: "

    gglobs.AudioLastCps     = 0
    gglobs.AudioLastCpm     = 0
    cps_count               = 0
    llimit                  = gglobs.AudioThreshold / 100 * gglobs.AudioPulseMax
    ilimit                  = 10000                             # the |pulse| must NOT exceed this value on first count in order to be countesd
                                                                # higher values give MORE counts!

    gglobs.AudioLast60CPS   = np.full(60, 0)                    # storing last 60 sec of CPS values
    gglobs.AudioMultiPulses = np.array([0])                     # time courses for last ~40 of the pulses
    gglobs.AudioRecording   = np.array([0])
    nandata                 = np.full(10, gglobs.NAN)           # creating a gap of 10 values
    chunks40                = (gglobs.AudioChunk + 10) * 40     # the length of 40 single pulses with gaps

    cstart                  = time.time()                       # to record the 1 sec collection period

    try:
        CHUNKstream = sd.InputStream(blocksize=gglobs.AudioChunk)   # no callback so we get blocking read
        CHUNKstream.start()
    except Exception as e:
        msg = "Audio cannot be activated. Is an input device missing?"
        #efprint(msg)
        exceptPrint(e, msg)
        gglobs.AudioThreadStop = True
        return

    while not gglobs.AudioThreadStop:
        #~wstart = time.time()
        try:
            record, overflowed = CHUNKstream.read(gglobs.AudioChunk)
            npdata = record.reshape(-1)     # convert from '[ [1] [0] [3] ...[2] ]'  to '[1, 0, 3, ..., 2]'
            gglobs.AudioRecording = np.concatenate((gglobs.AudioRecording, npdata))[-gglobs.AudioRate:]
        except Exception as e:
            info = fncname + "Exception reading stream "
            exceptPrint(e, info)
            npdata = np.array([0])
        #~print("npdata:  ", type(npdata), npdata[:10])
        #~wdt = (time.time() - wstart) * 1000

        #~xstart = time.time()
        if overflowed:
            # Input overflow.
            # In a stream opened with blocksize=0, indicates that data prior to
            # the first sample of the input buffer was discarded due to an overflow,
            # possibly because the stream callback is using too much CPU time.
            #
            # In a stream opened with a non-zero blocksize, it indicates that
            # data prior to one or more samples in the input buffer was discarded.
            # This can happen in full-duplex and input-only streams (including
            # playrec() and rec()).
            pass
            dprint("'overflowed' HAPPENED: Stream calls: record len:{}, overflowed:{}".format(len(record), overflowed) )

        # Determine if there is a count
        # ignore records catching pulses when already down; they were probably
        # already counted during last record
        gotCount = False
        if gglobs.AudioPulseDir:                        # True -> positive pulse
            if npdata[0] < +ilimit and npdata.max() > +llimit:
                gotCount = True
        else:                                           # False -> negative pulse
            if npdata[0] > -ilimit and npdata.min() < -llimit:
                gotCount = True

        if gotCount:
            cps_count  += 1
            gglobs.AudioMultiPulses   = np.concatenate((gglobs.AudioMultiPulses, nandata, npdata))[-chunks40:]
            #~print("AudioMultiPulses: ", len(gglobs.AudioMultiPulses), gglobs.AudioMultiPulses)
            #print("got count: Stream calls: {:7.3f}ms, record len:{}, overflowed:{}".format(dt, len(record), overflowed) )

        deltat = time.time() - cstart
        if deltat >= 1:
            gglobs.AudioLastCps   = cps_count / deltat  * 1.05 # correct for calc time
            gglobs.AudioLast60CPS = np.append(gglobs.AudioLast60CPS, gglobs.AudioLastCps)[1:]
            cps_count             = 0
            cstart                = time.time()

        #~xdt = (time.time() - xstart) * 1000
        #print("wdt:{:0.3f}ms, xdt:{:0.3f}ms".format(wdt, xdt))

    CHUNKstream.stop()
    CHUNKstream.close()


def getAudioCounterInfo(extended = False):
    """Info on settings of the sounddev thread"""

    if not gglobs.AudioConnection:   return "No connected device"

    AudioInfo = """Connected Device:             '{}'
Configured Variables:         {}
Geiger Tube Sensitivity:      {:0.1f} CPM/(µSv/h) ({:0.4f} µSv/h/CPM)
""".format(
                        gglobs.AudioDeviceName,
                        gglobs.AudioVariables,
                        gglobs.AudioSensitivity, 1 / gglobs.AudioSensitivity
                       )

    if extended == True:
        AudioInfo += """
Sampling Details:
- DEVICE:                     Input:'{}', Output:'{}'
- FORMAT:                     Input:'{}', Output:'{}'
- CHANNELS (1=Mono,2=Stereo): Input:'{}', Output:'{}'
- LATENCY [sec]:              Input:'{}', Output:'{}'
- Host API Index              {}
- RATE:                       {} (Samples per second)
- CHUNK:                      {} (Samples per read)
- Pulse Height Max            {} (System reported max signal)
- Pulse Direction             {} (negative or positive)
- Pulse Threshold             {}% of Pulse Height Max to trigger count"""\
                .format(
                        sd.default.device[0],
                        sd.default.device[1],
                        sd.default.dtype[0],
                        sd.default.dtype[1],
                        sd.default.channels[0],
                        sd.default.channels[1],
                        sd.default.latency[0],
                        sd.default.latency[1],
                        sd.default.hostapi,
                        sd.default.samplerate,

                        gglobs.AudioChunk,
                        gglobs.AudioPulseMax,
                        "POSITIVE" if gglobs.AudioPulseDir else "NEGATIVE",
                        gglobs.AudioThreshold,
                       )
        AudioInfo += "\nFull Device Query:\n"
        AudioInfo += str(sd.query_devices())
        AudioInfo += "\n"

    return AudioInfo


def toggleAudioCounterPulseDir():
    """toggle which pulse direction is sensed"""

    gglobs.AudioPulseDir = not gglobs.AudioPulseDir

    fprint(header("Switching Pulse Direction of AudioCounter"))
    if gglobs.AudioPulseDir:  direction = "POSITIVE"
    else:                     direction = "NEGATIVE"
    gglobs.AudioMultiPulses = gglobs.AudioMultiPulses[-10:]

    fprint("New pulse direction setting: {}".format(direction))


def getAudioValues(varlist):
    """Read all sounddev data; return empty dict when not available"""

    fncname   = "getAudioValues: "
    alldata = {}

    for vname in varlist:
        if   vname in ("CPM", "CPM1st", "CPM2nd", "CPM3rd"):
            cpm             = int(round(np.sum(gglobs.AudioLast60CPS), 0))
            cpm             = scaleVarValues(vname, cpm, gglobs.ValueScale[vname])
            alldata.update(  {vname: cpm})

        elif vname in ("CPS", "CPS1st", "CPS2nd", "CPS3rd"):
            cps             = int(round(gglobs.AudioLastCps, 0))
            cps             = scaleVarValues(vname, cps, gglobs.ValueScale[vname] )
            alldata.update(  {vname: cps})

    printLoggedValues(fncname, varlist, alldata)

    return alldata


def printAudioDevInfo(extended=False):
    """prints basic info on the AudioCounter device"""

    setBusyCursor()

    txt = "AudioCounter Device"
    if extended:  txt += " Extended"
    fprint(header(txt))
    fprint("Configured Connection:", "Default Audio Input")
    fprint(getAudioCounterInfo(extended=extended))

    setNormalCursor()


def plotAudio(dtype="Single Pulse", duration=None):
    """Plotting an audio plot"""

    fncname = "plotAudio: "

    data = gglobs.AudioPlotData
    #print(fncname + "dtype:{}, data:{}".format(dtype, data))

    subTitle = dtype

    if   dtype == "Single Pulse":
        pass

    elif dtype == "Multi Pulse":
        count     = len(data) / (gglobs.AudioChunk + 10) # 10 is: nan values as gap
        subTitle += " - {:1.0f} pulses shown".format(count)

    elif dtype == "Recording":
        subTitle += " - {:1.0f} second".format(duration)


    # Plotstyle
    plotstyle        = {'color'             : 'black',
                        'linestyle'         : 'solid',
                        'linewidth'         : '0.5',
                        'label'             : 'my label',
                        'markeredgecolor'   : 'black',
                        'marker'            : 'o',
                        'markersize'        : '1',
                        'alpha'             : 1,
                        }
    plt.close(99)
    fig2 = plt.figure(99, facecolor = "#E7cccc", dpi=gglobs.hidpiScaleMPL)
    wprint(fncname + "open figs count: {}, current fig: #{}".format(len(plt.get_fignums()), plt.gcf().number))

    plt.title("AudioCounter Device\n", fontsize=14, loc='center')
    plt.title("Audio Input", fontsize=10, fontweight='normal', loc = 'left')
    plt.title(subTitle, fontsize=10, fontweight='normal', loc = 'right')

    plt.grid(True)
    plt.subplots_adjust(hspace=None, wspace=.2 , left=.11, top=0.85, bottom=0.15, right=.97)
    plt.ticklabel_format(useOffset=False)

    # hide the cursor position from showing in the Nav toolbar
    ax1 = plt.gca()
    ax1.format_coord = lambda x, y: ""

    # canvas - this is the Canvas Widget that displays the `figure`
    # it takes the `figure` instance as a parameter to __init__
    canvas2 = FigureCanvas(fig2)
    canvas2.setFixedSize(1000, 550)
    navtoolbar = NavigationToolbar(canvas2, gglobs.exgg)

    labout = QTextBrowser()                  # label to hold the description
    labout.setFont(gglobs.fontstd)           # my std font for easy formatting
    labout.setText("")

    mtext1  =   "Device:           Input:'{}', Output:'{}'" .format(sd.default.device[0], sd.default.device[1])
    mtext1 += "\nSample Format:    {}"                      .format(gglobs.AudioFormat)
    mtext1 += "\nSampled Channels: {} (1=Mono, 2=Stereo)"   .format(gglobs.AudioChannels)
    mtext1 += "\nSampling rate:    {} per second"           .format(gglobs.AudioRate)
    mtext1 += "\nSamples:          {} per read"             .format(gglobs.AudioChunk)
    mtext1 += "\nPulse Height Max: ±{} "                    .format(gglobs.AudioPulseMax)
    mtext1 += "\nPulse Threshold:  {} %"                    .format(gglobs.AudioThreshold)
    mtext1 += "\nPulse Direction:  {} "                     .format(("POSITIVE" if gglobs.AudioPulseDir else "NEGATIVE"))

    if   dtype == "Single Pulse":
        labout.append("Showing the last detected AudioCounter pulse")

    elif dtype == "Multi Pulse":
        labout.append("Showing the last (up to 40) detected AudioCounter pulses")

    elif dtype == "Recording":
        labout.append("Showing a straight recording of {} second - no pulse detection applied".format(duration))

    labout.append(mtext1)

    d = QDialog()
    gglobs.plotAudioPointer = d
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setWindowTitle("AudioCounter Device")
    #d.setMinimumHeight(700)
    #d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    d.setWindowModality(Qt.WindowModal)

    okButton = QPushButton("OK")
    #okButton.setCheckable(True)
    okButton.setAutoDefault(True)
    okButton.clicked.connect(lambda:  d.done(0))

    singleButton = QPushButton("Single Pulse")
    #singleButton.setCheckable(True)
    singleButton.setAutoDefault(False)
    singleButton.clicked.connect(lambda: reloadAudioData("Single Pulse"))

    multiButton = QPushButton("Multi Pulse")
    #multiButton.setCheckable(True)
    multiButton.setAutoDefault(False)
    multiButton.clicked.connect(lambda: reloadAudioData("Multi Pulse"))

    recordButton = QPushButton("Recording")
    #recordButton.setCheckable(True)
    recordButton.setAutoDefault(False)
    recordButton.clicked.connect(lambda: reloadAudioData("Recording"))

    togglePulseDirButton = QPushButton("Toggle Pulse Direction")
    #togglePulseDirButton.setCheckable(True)
    togglePulseDirButton.setAutoDefault(False)
    togglePulseDirButton.clicked.connect(lambda: reloadAudioData("Toggle"))

    bbox    = QDialogButtonBox()
    bbox.addButton(okButton,                QDialogButtonBox.ActionRole)
    bbox.addButton(recordButton,            QDialogButtonBox.ActionRole)
    bbox.addButton(singleButton,            QDialogButtonBox.ActionRole)
    bbox.addButton(multiButton,             QDialogButtonBox.ActionRole)
    bbox.addButton(togglePulseDirButton,    QDialogButtonBox.ActionRole)

    layoutH = QHBoxLayout()
    layoutH.addWidget(bbox)
    layoutH.addWidget(navtoolbar)
    layoutH.addStretch()

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(layoutH)
    layoutV.addWidget(canvas2)
    layoutV.addWidget(labout)

    plt.xlabel("Time [millisec]", fontsize=14)
    plt.ylabel("Amplitude [rel]", fontsize=14)
    plt.ylim(-35000, 35000)

    x = np.arange(len(data)) * 1 / gglobs.AudioRate * 1000 # create the x-axis data
    plt.plot(x, data,  **plotstyle)
    #d.update() # "Updates the widget unless updates are disabled or the widget is hidden."

    fig2.canvas.draw_idle()
    playWav("ok")
    d.exec_()


def reloadAudioData(dtype):
    """closes the dialog and reopens"""

    #print("reloadAudioData: ", dtype)

    try: # fails if it had never been opened; but then it is closed anyway
        gglobs.plotAudioPointer.close()
    except:
        pass

    # when a recording was last, then gglobs.AudioMultiPulses will be 44100 samples long,
    # but even Multi needs only 40*(single pulse + 10) => 1680. Therfore crop it,
    # otherwise next Multi will show samples from recording
    if len(gglobs.AudioMultiPulses) > (40 * (gglobs.AudioChunk + 10)):
        #~gglobs.AudioMultiPulses = gglobs.AudioMultiPulses[-(gglobs.AudioChunk + 10):]
        gglobs.AudioMultiPulses =  np.array([0])

    duration = None

    if   dtype == "Single Pulse":
        gglobs.AudioPlotData = gglobs.AudioMultiPulses[-(gglobs.AudioChunk):] # last pulse only without nan
        #gglobs.AudioMultiPulses = gglobs.AudioMultiPulses[-(gglobs.AudioChunk):] # last pulse only without nan

    elif dtype == "Multi Pulse":
        pass    # will show all, i.e. up to last 40 pulses
        gglobs.AudioPlotData = gglobs.AudioMultiPulses

    elif dtype == "Recording":
        setBusyCursor()
        duration = 1 # seconds
        #gdev_audio.getLongChunk(duration)
        gglobs.AudioPlotData = gglobs.AudioRecording
        setNormalCursor()

    else: # dtype == "Toggle":
        #~gdev_audio.toggleAudioCounterPulseDir()
        toggleAudioCounterPulseDir()
        dtype = "Recording"
        duration = 1 # seconds
        #~gglobs.AudioPlotData = gglobs.AudioMultiPulses[-(gglobs.AudioChunk):] # last pulse only without nan
        gglobs.AudioMultiPulses = np.array([0]) # set to empty as old pulses do not make sense
        gglobs.AudioPlotData    = gglobs.AudioRecording

    plotAudio(dtype, duration)


def audio_callback(indata, frames, time, status):
    """This is called (from a separate thread) for each audio block."""

    global liveq

    fncname = "audio_callback: "
    #print(fncname + "np.shape(indata): ", np.shape(indata), ", frames: ", frames, ", time: ", time, ", status: ", status)

    if status:
        edprint(fncname + "status: ", status)

    liveq.put(indata[:, 0])


def updateAnimationPlot(frame):
    """This is called by matplotlib for each animated plot update.

    Typically, audio callbacks happen more frequently than plot updates,
    therefore the queue tends to contain multiple blocks of audio data.
    """
    global liveq, plotdata

    start = time.time()

    fncname = "updateAnimationPlot: "

    #print(fncname + "frame", frame)

    #counter = 0
    while True:
        try:
            data = liveq.get_nowait()
            #print("liveq: ", counter)
            #counter += 1
        except queue.Empty:
            #edprint("queue.Empty: ", queue.Empty)
            break

        shift    = len(data)
        plotdata = np.roll(plotdata, -shift)
        plotdata[-shift:] = data

        lines[0].set_ydata(plotdata)

    return lines


def makeEnde():
    global ende
    ende = True


def showLiveAudioSignal():
    """Plotting a rolling audio plot"""

    fncname = "showLiveAudioSignal: "

    global liveq, lines, plotdata, ende
    global args_device, args_channels, args_samplerate

    liveq = queue.Queue()
    ende  = False

    try:
        #print(fncname + "Audio Devices found on system:\n{}".format(sd.query_devices()))

        #~args_samplerate = 44100 # get default instead
        args_samplerate = None
        #args_samplerate = 11050*2

        args_window     = 20000
        args_downsample = 1
        #~args_channels   = (1, ) # needed as tuple
        args_channels   = 1
        #~args_interval   = 300
        args_interval   = 50
        args_device     = None

        if args_samplerate is None:
            device_info = sd.query_devices(args_device, 'input')
            args_samplerate = device_info['default_samplerate']
        #print(fncname + "parameters: ", args_samplerate, args_window, args_downsample, args_channels, args_interval, args_device)

        length   = int(args_window * args_samplerate / (1000 * args_downsample)) # time length in ms
        plotdata = np.zeros((length))                        # create buffer of required length
        #~print("length: ", length)                                                # 20000 * 44100 / (1000 * 1) = 882000
        #~print("plotdata: ", plotdata)                                            # [[0.] [0.] [0.] ... [0.] [0.] [0.]]
        #~print("np.shape(plotdata): ", np.shape(plotdata))                        # np.shape(plotdata):  (441000, 1)

    except Exception as e:
        srcinfo = fncname + "Exception: "
        exceptPrint(e, srcinfo)

    plt.close(77)
    fig77 = plt.figure(77, facecolor = "#E7cccc", dpi=gglobs.hidpiScaleMPL) # light red
    vprint(fncname + "open figs count: {}, current fig: #{}".format(len(plt.get_fignums()), plt.gcf().number))

    lines = plt.plot(plotdata) # creates a plot array x=0,1,2,... y=0,0,0,

    #~for column, line in enumerate(lines):
        #~print("column: ", column, ", lines: ", lines, column, line)
        #~print("x: ", line.get_xdata(orig=True), "\ny: ", line.get_ydata(orig=True))

    plt.title("AudioCounter Device\n", fontsize=14, loc='center')
    plt.title("Live Audio Signal", fontsize=10, fontweight='normal', loc = 'left')

    plt.grid(True)
    plt.subplots_adjust(hspace=None, wspace=.2 , left=.11, top=0.85, bottom=0.15, right=.97)
    plt.ticklabel_format(useOffset=False)

    # do not draw x ticks and x labels
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,      # ticks along the bottom edge are off
        top=False,         # ticks along the top edge are off
        labelbottom=False) # labels along the bottom edge are off

    # hide the cursor position from showing in the Nav toolbar
    ax1 = plt.gca()
    ax1.format_coord = lambda x, y: ""

    # canvas - this is the Canvas Widget that displays the `figure`
    # it takes the `figure` instance as a parameter to __init__
    canvas77 = FigureCanvas(fig77)
    canvas77.setFixedSize(1000, 550)
    navtoolbar = NavigationToolbar(canvas77, gglobs.exgg)

    d = QDialog()
    gglobs.plotAudioPointer = d
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setWindowTitle("AudioCounter Device")
    #d.setMinimumHeight(700)
    #d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    d.setWindowModality(Qt.WindowModal)

    okButton = QPushButton("OK")
    okButton.setAutoDefault(True)
    okButton.clicked.connect(lambda:  makeEnde())

    bbox    = QDialogButtonBox()
    bbox.addButton(okButton,                QDialogButtonBox.ActionRole)

    layoutH = QHBoxLayout()
    layoutH.setAlignment(Qt.AlignLeft);
    layoutH.addWidget(bbox)
    layoutH.addWidget(navtoolbar)
    layoutH.addStretch() # all stretches needed, very odd!
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()
    layoutH.addStretch()

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(layoutH)
    layoutV.addWidget(canvas77)

    plt.xlabel("\nTime", fontsize=14)
    plt.ylabel("Amplitude [rel]", fontsize=14)
    plt.ylim(-35000, 35000)

    # https://stackoverflow.com/questions/16732379/stop-start-pause-in-python-matplotlib-animation
  #~ani = mplanim.FuncAnimation(fig77, updateAnimationPlot, interval=args_interval, blit=True)
    ani = mplanim.FuncAnimation(fig77, updateAnimationPlot, interval=args_interval, blit=False)

    playWav("ok")

    d.open()    # needed!

    start = time.time()
    while True: # one cycle is about 350 ms
        stream = sd.InputStream(device=args_device, channels=args_channels, samplerate=args_samplerate, callback=audio_callback)

        with stream: # not doing anything, but still needed!
            pass

        Qt_update() # ohne dies nur leerer Fensterrahmen
        #d.show() # kein Zusatznutzen; ohne Qt_update nur leeres Bild
        #time.sleep(0.2)
        print("time delta: {:0.3f} ms".format(1000*(time.time() -start))) # alle 300 - 400 ms
        start = time.time()

        if ende: break

    #print("with stream done")

    d.close()
    ani.event_source.stop() # ouff, this stops the animation!


def turnEia(tdir, mmovie):

    turnDir = tdir
    mmovie.stop()
    time.sleep(0.3)
    mmovie.start()


def showAudioEia():
    """https://www.gqelectronicsllc.com/forum/topic.asp?TOPIC_ID=9517&#7803"""

    fncname = "showAudioEia: "

    from PyQt5.QtGui import QMovie

    d = QDialog()

    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setWindowTitle("Eia")
    #d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    d.setWindowModality(Qt.WindowModal)

    okButton = QPushButton("OK")
    okButton.setAutoDefault(True)
    okButton.clicked.connect(lambda:  d.close())

    singleButton = QPushButton("Left Turn")
    singleButton.setAutoDefault(False)
    singleButton.clicked.connect(lambda: turnEia("Left", movie))

    rightButton = QPushButton("Right Turn")
    rightButton.setAutoDefault(False)
    rightButton.clicked.connect(lambda: turnEia("Right", movie))

    bbox    = QDialogButtonBox()
    bbox.addButton(okButton,                QDialogButtonBox.ActionRole)
    bbox.addButton(singleButton,            QDialogButtonBox.ActionRole)
    bbox.addButton(rightButton,             QDialogButtonBox.ActionRole)


    peia = "".join(map(chr, [101, 105, 97, 46, 103, 105, 102]))
    dp = os.path.join(getProgPath(), gglobs.gresDirectory, peia)

    label = QLabel()
    movie = QMovie(dp)
    label.setMovie(movie)
    movie.start()

    layoutH = QHBoxLayout()
    layoutH.addWidget(bbox)
    layoutH.addStretch()

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(layoutH)
    layoutV.addWidget(label)

    playWav("ok")

    d.exec()



##############################################################################
# following only relevant for use via 'main()'
# like: 'python3 geigerlog/gdev_audio.py 7'


def getLongChunk(duration, recno):
    """record duration seconds of audio"""

    print("#{} - Recording for {} sec".format(recno, duration))

    reclen = int(sd.default.samplerate * duration)
    mystream = sd.InputStream(blocksize=reclen)
    mystream.start()
    record, overflowed = mystream.read(reclen) # get one record
    gglobs.AudioMultiPulses = record.reshape(-1)
    mystream.stop()


def plotDataT(AudioMultiPulses, inversrate):
    """plot record of sound data"""

    if len(AudioMultiPulses) == 0: return

    # Set "QT_STYLE_OVERRIDE" to avoid this message:
    #   QApplication: invalid style override 'gtk' passed, ignoring it.
    #   Available styles: Windows, Fusion
    os.environ["QT_STYLE_OVERRIDE"] = ""

    fig = plt.figure("Total", figsize=(14, 6), dpi=gglobs.hidpiScaleMPL)

    show = 0    # 1: accumulate the counts (don't delete the figure)
                # 0: delete last figure, and show a fresh record every time
    if    show: pass
    else:       fig.clf()

    x = np.arange(len(gglobs.AudioMultiPulses)) * inversrate * 1000 # create X-axis values
    y = gglobs.AudioMultiPulses                                     # Y-axis values
    #~print("plotDataT: x    :", len(x), x)
    #~print("plotDataT: y    :", len(y), y)
    plt.plot(x, y,  'k.', linestyle='solid', linewidth=0.3, markersize=0.5)
    plt.xlabel("millisec")
    plt.ylim(-35000, 35000)
    plt.pause(.5)


def main():
    """
    Start with: 'python3 gdev_audio.py' to get a list of available Sound Devices
    Start with: 'python3 gdev_audio.py N' to get a list of available Sound Devices,
                followed by N recordings and recoding plots from the default device
    """

    print("Available Sound Devices:")
    print(sd.query_devices())
    print()

    # Setting the soundcard
    sd.default.reset()
    sd.default.device       = (None, None)              # selects the default device
    sd.default.latency      = (1.0, 1.0)                # selects the default device
    sd.default.dtype        = ('int16', 'int16')        # signed 16 bit resolution for both input and output
    sd.default.channels     = (1, 1)                    # Mono for both input and output
    sd.default.samplerate   = 44100                     # options: 96000, 48000, 44100, 22050, 11025

    print("Sound Driver Settings:\n    DEVICE:{}, CHANNELS:{}, FORMAT:{}, Latency:{}, RATE:{}, Host API Index:{}"\
                    .format(
                            sd.default.device,
                            sd.default.channels,
                            sd.default.dtype,
                            sd.default.latency,
                            sd.default.samplerate,
                            sd.default.hostapi,
                           )
         )

    inversrate              = 1 / sd.default.samplerate
    gglobs.AudioMultiPulses = np.zeros(256)

    try:
        recs = abs(int(sys.argv[1])) # if none given on command line, an exception will be raised
    except:
        recs = 0

    for i in range(0, recs):
        try:
            getLongChunk(1.0, i + 1)
            plotDataT(gglobs.AudioMultiPulses, inversrate)
        except Exception as e:
            print("gdev_audio.py: main: Exception :", e)

    if recs > 0:
        plotDataT(gglobs.AudioMultiPulses, inversrate)
        print("Done recording. To Exit close graph, or press CTRL-C in terminal")
        plt.show()


if __name__ == '__main__':
    import matplotlib.pyplot as plt # plotting possible only when run as main
    main()
    print()




#! /usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
AmbioMon++ - calling AmbioMon++ running on ESP32
"""

###############################################################################
#    This file is part of GeigerLog.
#
#    GeigerLog is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GeigerLog is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GeigerLog.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

__author__          = "ullix"
__copyright__       = "Copyright 2016, 2017, 2018, 2019, 2020, 2021"
__credits__         = [""]
__license__         = "GPL3"

from   gsup_utils           import *
import gsup_sql                             # database handling


###############################################################################
# used only internally within gdev_ambiomon.py

def _getAmbioUrl():
    # Port is always 80
    return "http://{}".format(gglobs.AmbioServerIP)


def _getAmbioDevice():
    # get the device like 'ESP32-WROOM-Dev' by calling page '/amID' yielding:
    #   2020-06-02 09:39:01, ESP32-WROOM-Dev, 10.0.0.85, ambiomon.local, 3

    try:
        url   = _getAmbioUrl() + "/amid"
        page  = urllib.request.urlopen(url, timeout=gglobs.AmbioTimeout)
        data  = page.read().decode("UTF-8").split(",")
    except Exception as e:
        edprint("Failed _getAmbioDevice at url:'{}' with Exception: ".format(url), e, debug=True)
        efprint ("{} Getting AmbioMon Device Identifier @ url:<br>'{}' failed with exception:<br>{}".format(stime(), url, cleanHTML(e)))
        return "NONE"

    return data[1].strip()


def _convertUnixtimeToDateTime(unixtime):
    """convert a Unix time in sec into datetime like: 2020-05-10 11:48:50"""

    return datetime.datetime.utcfromtimestamp(unixtime).strftime('%Y-%m-%d %H:%M:%S')


def _printException(e, rec):
    dprint  ("Exception _parseValueAdder: ", e)
    efprint ("AmbioMon History Download had error: ", cleanHTML(e))
    qefprint("Record is ignored, has Error: ", rec)
    dprint  ("Record is ignored, has Error: ", rec)


def _parseValueAdderCPS(i, rec):
    """Add the parse results to the *.his list"""

    # create the data for the database; datalist covers:
    # Index, DateTime, <modifier>,  CPM, CPS, CPM1st, CPS1st, CPM2nd, CPS2nd,  CPM3rd, CPS3rd, Temp, Press, Humid, X
    # 0      1         2            3    4    5       6       7        8       9       10      11    12     13     14

    #~ print("_parseValueAdder: i, rec:", i, rec)
    try:
        reclist      = rec.split(",", 5)        #
        datalist     = [None] * (gglobs.datacolsDefault + 2)   # (13 + 2)=15 x None
        datalist[0]  = i                        # byte index
        datalist[1]  = reclist[0]               # DateTime UTC
        datalist[2]  = "0 hours"                # the modifier for julianday; here: no modification
        datalist[4]  = reclist[1]               # cps

        datalist[9]  = reclist[2]               # selector
        datalist[10] = reclist[3]               # volt

        gglobs.HistoryDataList.append (datalist)

    except Exception as e:
        _printException(e, rec)


def _parseValueAdderCAM(i, rec):
    """Add the parse results to the *.his list"""

    # create the data for the database; datalist covers:
    # Index, DateTime, <modifier>,  CPM, CPS, CPM1st, CPS1st, CPM2nd, CPS2nd,  CPM3rd, CPS3rd, Temp, Press, Humid, X
    # 0      1         2            3    4    5       6       7        8       9       10      11    12     13     14

    #~ print("_parseValueAdder: i, rec:", i, rec)
    try:
        reclist      = rec.split(",", 9)    # results in datetime + 6 values + 1 extra for the rest
        datalist     = [None] * (gglobs.datacolsDefault + 2)   # (13 + 2)=15 x None
        datalist[0]  = i                    # byte index
        datalist[1]  = reclist[0]           # DateTime UTC
        datalist[2]  = "0 hours"            # the modifier for julianday; here: no modification
        datalist[3]  = reclist[1]           # cpm

        datalist[9]  = reclist[6]           # selector
        datalist[10] = reclist[7]           # volt

        datalist[11] = reclist[2]           # temp
        datalist[12] = reclist[3]           # pressure
        datalist[13] = reclist[4]           # humidity
        datalist[14] = reclist[5]           # air quality

        gglobs.HistoryDataList.append (datalist)

    except Exception as e:
        _printException(e, rec)


# gdev_ambiomon.py use only
###############################################################################


###############################################################################
# Public Functions


def makeAmbioHistory(source, device):  # source == "AMDeviceCAM" or "AMDeviceCPS"
    """read the history from an AmbioMon device"""

    fncname = "makeAmbioHistory: "

    if not gglobs.AmbioActivation:
        emsg = "AmbioMon device is not activated"
        dprint(fncname + emsg)
        return -1, emsg

    dprint(fncname + "from source: '{}' using device: '{}'".format(source, device))
    dprint(fncname + "gglobs.AMFilePath: {}".format(gglobs.AMFilePath))
    setDebugIndent(1)

    error           = 0         # default: no error
    message         = ""        # default: no message

    start = time.time()

    if   source == "AMDeviceCPS" or source == "AMDeviceCAM":
    # get data from AmbioMon device

        if not gglobs.AmbioConnection:
            emsg = "AmbioMon device is not connected"
            dprint(fncname + emsg)
            return -1, emsg

        if   source == "AMDeviceCPS": url = _getAmbioUrl() + "/log.cps"
        elif source == "AMDeviceCAM": url = _getAmbioUrl() + "/log.cam"
        fprint("Using URL: ", url)

        try:
            page        = urllib.request.urlopen(url, timeout=10)
            devicedata  = page.read()   # type(devicedata) = <class 'byte'>
        except Exception as e:
            dprint("Failed urlib: Exception: ", e, debug=True)
            playWav("error")
            efprint("{} Reading data from Web failed with exception:<br>{}".format(stime(), cleanHTML(e)))
            setDebugIndent(0)
            Qt_update()
            return -1, cleanHTML(e)

    elif source == "AMFileCPS" or source == "AMFileCAM":
    # get data from AmbioMon File
        try:
            devicedata = readBinaryFile(gglobs.AMFilePath)
        except Exception as e:
            playWav("error")
            efprint("{} Reading data from File failed with exception:<br>{}".format(stime(), cleanHTML(e)))
            setDebugIndent(0)
            Qt_update()
            return -1, cleanHTML(e)

    stop       = time.time()

    dtime      = (stop - start)
    lendevicedata = len(devicedata)
    msg        = "Downloaded {:0.1f} kB from {} in {:0.1f} milliseconds ({:0.1f} kB/s)".format(lendevicedata / 1000, source, dtime * 1000, lendevicedata/dtime/1000)
    vprint(msg)
    fprint(msg)
    fprint("Processing data ...")
    Qt_update()

    data_originDB   = (stime(), gglobs.AmbioDeviceDetected)
    dbheader        = "File created by reading log from device"
    dborigin        = "Download from device"
    dbdevice        = "{}".format(data_originDB[1])

    dumpdata = []
    if   source == "AMDeviceCPS" or source == "AMFileCPS":
        #~ upformat = "<HBBL" + "H" * 60
        #~upformat = "<BBBL" + "H" * 60
        upformat = "<BBHL" + "H" * 60
        wprint("Unpack Format: ", upformat)
        for i in range(0, len(devicedata), 128):
            dbuf = devicedata[i : i + 128]
            data = struct.unpack(upformat, dbuf)
            #~volt = data[2] + 350     # volt is given minus 350 to fit into 1 byte
            volt = data[2]            # volt is given as int
            dcfg = data[1] & 0x03    # selector is 1, or 2, or 3

            for j in range(0, 60):   # make 60 records
                dt = _convertUnixtimeToDateTime(data[3] - 60 + j)
                sdata = "{},{},{},{}".format(
                                    dt,
                                    data[4 + j],
                                    dcfg,
                                    volt,
                                    )
                dumpdata.append(sdata)
            #~ if i > 1000: break

    elif source == "AMDeviceCAM" or source == "AMFileCAM":
        upformat = "<BBHLLfffff"
        wprint("Unpack Format: ", upformat)
        for i in range(0, len(devicedata), 32):
            dbuf = devicedata[i : i + 32]
            data = struct.unpack(upformat, dbuf)
            #~volt = data[2] + 350     # volt is given minus 350 to fit into 1 byte
            volt = data[2]           # volt is now in 2 bytes, int 16
            dcfg = data[1] & 0x03    # selector is 1, or 2, or 3

            sdata = "{},{},{:0.2f},{:0.2f},{:0.2f},{:0.2f},{},{}".format( # make single record
                                _convertUnixtimeToDateTime(data[3]),
                                data[4],
                                data[5],
                                data[6],
                                data[7],
                                data[8],
                                dcfg,
                                volt,
                                )
            dumpdata.append(sdata)

    #~for a in dumpdata[:6]:  print("dumpdata:", a)
    #~print()
    #~for a in dumpdata[-6:]: print("dumpdata:", a)

    if len(dumpdata) == 0:
        error   = -1
        message = "No valid data found!"
        dprint(message)

    else:
    # parse HIST data
        gglobs.HistoryDataList      = []
        gglobs.HistoryParseList     = []
        gglobs.HistoryCommentList   = []

        #~ _getParsedHistoryCPS(dumpdata)
        if   source == "AMDeviceCPS" or source == "AMFileCPS":
            for i, rec in enumerate(dumpdata):
                _parseValueAdderCPS(i, rec)

        elif source == "AMDeviceCAM" or source == "AMFileCAM":
            for i, rec in enumerate(dumpdata):
                _parseValueAdderCAM(i, rec)

    # add headers
        dbhisClines    = [None] * 3
        #                  ctype    jday, jday modifier to use time unmodified
        dbhisClines[0] = ["HEADER", None, "0 hours", dbheader]
        dbhisClines[1] = ["ORIGIN", None, "0 hours", dborigin]
        dbhisClines[2] = ["DEVICE", None, "0 hours", dbdevice]

    # write to database
        gsup_sql.DB_insertDevice        (gglobs.hisConn, *data_originDB)
        gsup_sql.DB_insertComments      (gglobs.hisConn, dbhisClines)
        gsup_sql.DB_insertComments      (gglobs.hisConn, gglobs.HistoryCommentList)
        gsup_sql.DB_insertData          (gglobs.hisConn, gglobs.HistoryDataList)
        gsup_sql.DB_insertParse         (gglobs.hisConn, gglobs.HistoryParseList)

    # write device data to database
        saveToBin = True
        if saveToBin:   gsup_sql.DB_insertBin(gglobs.hisConn, devicedata)

        fprint("Database is created", debug=gglobs.debug)

    setDebugIndent(0)
    return error, message


def setAmbioServerIP():
    """Set settings for Ambiomon when in Station mode"""

    # setting a server address must be possible without a prior connection,
    # because Init fails if IP not accessible

    fncname = "setAmbioServerIP: "

    dprint(fncname)
    setDebugIndent(1)

    # AmbioServer IP
    lstaip = QLabel("Domain Name or IP Adress of AmbioMon")
    lstaip.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
    staip  = QLineEdit()
    staip.setToolTip("Enter the local Domain Name (if supported) or local IP where your AmbioMon device can be reached")
    staip.setText("{}".format(gglobs.AmbioServerIP))

    graphOptions=QGridLayout()
    graphOptions.setContentsMargins(10,10,10,10) # spacing around the graph options
    graphOptions.addWidget(lstaip, 0, 0)
    graphOptions.addWidget(staip,  0, 1)

    # Dialog box
    d = QDialog()
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("Set AmbioMon Device IP")
    d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    #d.setWindowModality(Qt.WindowModal)
    d.setMinimumWidth(400)

    # Buttons
    bbox = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok )
    bbox.accepted.connect(lambda: d.done(1))
    bbox.rejected.connect(lambda: d.done(-1))

    gglobs.btn = bbox.button(QDialogButtonBox.Ok)
    gglobs.btn.setEnabled(True)

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(graphOptions)
    layoutV.addWidget(bbox)

    if gglobs.logging:         # no change of parameters when logging
        gglobs.btn  .setEnabled(False)
        staip       .setEnabled(False)

    retval = d.exec()
    #print("reval:", retval)

    if retval != 1:
        # ESCAPE pressed or Cancel Button
        dprint(fncname + "Canceling; settings unchanged")

    else:
        # OK pressed
        gglobs.AmbioServerIP = staip.text().strip()

        dprint(fncname + "new IP Address:    ", gglobs.AmbioServerIP)
        fprint(header("Configure AmbioMon WiFi Station Mode"))
        fprint("AmbioMon Station Mode IP:"     , "{}"    .format(gglobs.AmbioServerIP))

    setDebugIndent(0)


def setAmbioLogDatatype():
    """Set the datatype GeigerLog will request from Ambiomon (LAST or AVG)"""

    if not gglobs.AmbioConnection: return "No connected device"

    dprint("setAmbioLogDatatype:")
    setDebugIndent(1)

    lstaip = QLabel("Data as LAST value or last AVG value?\n(No change during logging)" + " "*15)
    lstaip.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)

    b0 = QRadioButton("LAST")             # this is default
    b1 = QRadioButton("AVG")
    b0.setToolTip("Select 'LAST' to get last value of data, or 'AVG' to get last 1 min average of values")
    b1.setToolTip("Select 'LAST' to get last value of data, or 'AVG' to get last 1 min average of values")

    if gglobs.AmbioDataType == "LAST" : b0.setChecked(True)
    else:                               b1.setChecked(True)

    layoutT = QHBoxLayout()
    layoutT.addWidget(b0)
    layoutT.addWidget(b1)

    graphOptions=QGridLayout()
    graphOptions.setContentsMargins(10,10,10,10) # spacing around the options
    graphOptions.addWidget(lstaip,          0, 0)
    graphOptions.addLayout(layoutT,         0, 1)

    # Dialog box
    d = QDialog()
    d.setWindowIcon(gglobs.iconGeigerLog)
    d.setFont(gglobs.fontstd)
    d.setWindowTitle("Select Data Type Mode")
    d.setWindowModality(Qt.ApplicationModal)
    #d.setWindowModality(Qt.NonModal)
    #d.setWindowModality(Qt.WindowModal)
    d.setMinimumWidth(400)

    # Buttons
    bbox = QDialogButtonBox()
    bbox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok )
    bbox.accepted.connect(lambda: d.done(1))
    bbox.rejected.connect(lambda: d.done(-1))

    gglobs.btn = bbox.button(QDialogButtonBox.Ok)
    gglobs.btn.setEnabled(True)

    layoutV = QVBoxLayout(d)
    layoutV.addLayout(graphOptions)
    layoutV.addWidget(bbox)

    if gglobs.logging:         # no change of parameters when logging
        gglobs.btn  .setEnabled(False)

    retval = d.exec()
    #print("reval:", retval)

    if retval != 1:
        # ESCAPE pressed or Cancel Button clicked
        dprint("setAmbioLogDatatype: Settings unchanged")
    else:
        if b0.isChecked(): gglobs.AmbioDataType = "LAST"    # b0.isChecked()
        else:              gglobs.AmbioDataType = "AVG"     # b1.isChecked()

        fprint(header("Select Data Type Mode"))
        fprint("AmbioMon Data Type:", "{}"    .format(gglobs.AmbioDataType), debug=True)

    setDebugIndent(0)


def getAmbioValues(varlist):
    """Read all AmbioMon data"""

    fncname = "getAmbioValues: "

    alldata = {}

    start = time.time()
    attempts     = 0        # no of attempts to get a connection
    attempts_max = 5        # max no of attempts

    while attempts < attempts_max:
        try:
            if not gglobs.logging:  break
            attempts += 1

            if gglobs.AmbioDataType == "LAST":  url = _getAmbioUrl() + "/lastdata"
            else:                               url = _getAmbioUrl() + "/lastavg"

            page  = urllib.request.urlopen(url, timeout=gglobs.AmbioTimeout)        # 40...120 ms, rest < 0.05 ms
            data  = page.read().decode("UTF-8").split(",")
            #~print("page: geturl : ", page.geturl())
            #~print("page: getcode: ", page.getcode())
            #~for a, b in page.info().items(): print ("page.info:  {:20s}   {}".format(a, b))
            #~print(fncname + "page.read: ", data)

            #stop = time.time()
            #~start = time.time()

            for a in varlist:
                if   a == "CPM":
                                    cpm = float(data[1]) if (float(data[1]) != 4294967295) else gglobs.NAN
                                    alldata.update({"CPM":      cpm                 })   # CPM
                elif a == "CPS":    alldata.update({"CPS":      float(data[2])      })   # CPS
                elif a == "CPM1st": alldata.update({"CPM1st":   float(data[6])      })   # BME680 resistance kOhm
                elif a == "CPS1st": alldata.update({"CPS1st":   float(data[6])      })   # BME680 resistance kOhm
                elif a == "CPM2nd": alldata.update({"CPM2nd":   float(data[9])      })   # Chip Voltage
                elif a == "CPS2nd": alldata.update({"CPS2nd":   float(data[10])     })   # Supply Voltage
                elif a == "CPM3rd": alldata.update({"CPM3rd":   float(data[9])      })   # Chip Voltage
                elif a == "CPS3rd": alldata.update({"CPS3rd":   float(data[10])     })   # Supply Voltage
                elif a == "T":      alldata.update({"T":        float(data[3])      })   # T     from BMEX80
                elif a == "P":      alldata.update({"P":        float(data[4])      })   # P     from BMEX80
                elif a == "H":      alldata.update({"H":        float(data[5])      })   # H     from BMEX80
                elif a == "X":      alldata.update({"X":        float(data[6])      })   # Airq  BME680 resistance kOhm (plotted as (LOG10(1/VAL) + 3 )*50)

            break

        except Exception as e:
            msg = fncname + "#{}: {}: Reading '{}' failed <br>with exception:  '{}'".format(attempts, stime(), url, cleanHTML(e))
            exceptPrint(e, msg)
            if attempts == attempts_max:  playWav("error")
            qefprint(msg)
            qefprint(" Timeout setting: {} sec".format(gglobs.AmbioTimeout))
            Qt_update()
            time.sleep(2) # does a pause help? 0.5 not much; 1 not really

    stop = time.time()
    loadTime = round((stop - start) * 1000, 2)      # ms
    alldata.update({"CPM1st":   float(loadTime)})   # Overall duration of loading the data set

    printLoggedValues(fncname, varlist, alldata)

    return alldata


def initAmbioMon():
    """Initialize AmbioMon"""

    fncname = "initAmbioMon: "
    errmsg  = ""

    if not gglobs.AmbioActivation:
        errmsg = "Initialzing AmbioMon not possible as AmbioMon device is not activated in configuration"
        dprint(fncname + errmsg)
        return errmsg

    dprint(fncname + "Initialzing AmbioMon")
    setDebugIndent(1)

    # set configuration
    if gglobs.AmbioServerIP     == "auto":  gglobs.AmbioServerIP     = "192.168.4.1"
    if gglobs.AmbioTimeout      == "auto":  gglobs.AmbioTimeout      = 3    # 1 sec scheint zu kurz, 5 sec bringt nichts
    if gglobs.AmbioCalibration  == "auto":  gglobs.AmbioCalibration  = 154
    if gglobs.AmbioVariables    == "auto":  gglobs.AmbioVariables    = "CPM, CPS, CPM1st, CPS1st, CPM2nd, CPS2nd, CPM3rd, CPS3rd, T, P, H, X"
    if gglobs.AmbioDataType     == "auto":  gglobs.AmbioDataType     = "LAST"           # using 'lastdata' call, not avg for 'lastavg'

    setCalibrations(gglobs.AmbioVariables, gglobs.AmbioCalibration)

    gglobs.AmbioDeviceName        = "AmbioMon++"
    gglobs.AmbioDeviceDetected    = _getAmbioDevice()            # expect like 'ESP32-WROOM-Dev'
    gglobs.Devices["AmbioMon"][0] = gglobs.AmbioDeviceDetected

    if gglobs.AmbioDeviceDetected == "NONE":
        errmsg = "<br>No connection to: AmbioServerIP: '{}'".format(gglobs.AmbioServerIP)
        edprint(fncname + errmsg)
        return errmsg

    # connected
    gglobs.AmbioConnection = True
    dprint(fncname + "connected to: AmbioServerIP: '{}', detected device: '{}'".format(gglobs.AmbioServerIP, gglobs.AmbioDeviceDetected))

    setLoggableVariables("AmbioMon", gglobs.AmbioVariables)

    setDebugIndent(0)
    return errmsg


def terminateAmbioMon():
    """opposit of init ;-)"""

    if not gglobs.AmbioConnection: return

    fncname = "terminateAmbioMon: "

    dprint(fncname)
    gglobs.AmbioConnection = False


def printAmbioInfo(extended=False):
    """prints basic info on the AmbioMon device"""

    setBusyCursor()

    txt = "AmbioMon Device"
    if extended:  txt += " Extended"
    fprint(header(txt))
    fprint("Configured Connection:", "Server IP: '{}'".format(gglobs.AmbioServerIP))
    fprint(getAmbioInfo(extended=extended))

    setNormalCursor()


def getAmbioInfo(extended=False):
    """Info on the AmbioMon Device"""

    if not gglobs.AmbioConnection: return "No connected device"

    AmbioInfo = """Connected Device:             '{}'
Configured Variables:         {}
Geiger Tube Sensitivity:      {:0.1f} CPM/(µSv/h) ({:0.4f} µSv/h/CPM)
Data Handling [LAST, AVG]:    {}
TESTING USAGE:  CPM         = GMtube CPM
                CPS         = GMtube CPS

                CPM1st      = FreeHeap [B]
                CPS1st      = AllocHeap [B]

                CPM2nd      = Batt Voltage ADS1115
                CPS2nd      = Batt Voltage ESP32

                CPM3rd      = Selector position
                CPS3rd      = Anode Voltage

                T           = BME680 Temp
                P           = BME680 Press
                H           = BME680 Humid
                X           = BME680 (log10(1/R[kOhm]) + 3 )*50

""".format(\
                                gglobs.AmbioDeviceDetected, \
                                gglobs.AmbioVariables, \
                                gglobs.AmbioCalibration, 1 / gglobs.AmbioCalibration, \
                                gglobs.AmbioDataType \
                                )
    if extended == True:
        AmbioInfo += """
Geiger tube voltage:          {} Volt
Cycletime:                    {} sec
Frequency:                    {} Hz
PWM:                          {}
Saving:                       {}
""".format(\
                                gglobs.AmbioVoltage,
                                gglobs.AmbioCycletime,
                                gglobs.AmbioFrequency,
                                gglobs.AmbioPwm,
                                gglobs.AmbioSav,
                                )
    return AmbioInfo

